/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

#undef Arc
#undef ArcSdirSorter
#undef ArcSorter
#undef ArcTdirSorter
#undef ArcTessellator
#undef Backend
#undef BasicCurveEvaluator
#undef BasicSurfaceEvaluator
#undef BezierArc
#undef Bin
#undef Breakpt
#undef Buffer
#undef CachingEvaluator
#undef Checker
#undef CoveAndTiler
#undef Curve
#undef Curvelist
#undef DisplayList
#undef Dlnode
#undef Edge
#undef Flist
#undef FlistSorter
#undef GridTrimVertex
#undef GridVertex
#undef Gridline
#undef Hashtable
#undef Hull
#undef Jarcloc
#undef JumpBuffer
#undef Knotspec
#undef Knotvector
#undef Mapdesc
#undef Maplist
#undef Mesher
#undef Monotonizer
#undef NurbsTessellator
#undef O_curve
#undef O_nurbscurve
#undef O_nurbssurface
#undef O_pwlcurve
#undef O_surface
#undef O_trim
#undef Patch
#undef Patchlist
#undef Patchspec
#undef Pool
#undef PooledObj
#undef Property
#undef Pspec
#undef PwlArc
#undef PwlArcPool
#undef Quilt
#undef Quiltspec
#undef Renderhints
#undef Slicer
#undef Sorter
#undef Splinespec
#undef Subdivider
#undef TrimRegion
#undef TrimVertex
#undef TrimVertexPool
#undef Trimline
#undef Uarray
#undef Varray
