/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

/////////////////////////////////////////////////////////////////////////////
//
// IfSorter class: takes a IfShapeList and sorts it (in place) by various
// properties. The resulting list is then easily examined to see which
// shapes can be processed together.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef  _IF_SORTER_
#define  _IF_SORTER_

#include "IfShapeList.h"

class IfShape;

class IfSorter {

  public:
    IfSorter();
    ~IfSorter();

    void	sort(IfShapeList &shapeList);

  private:
    static IfSorter	*curSorter;	// Saves "this" during qsort
    IfShapeList		*curShapeList;	// For verification

    // Compares 2 shapes for qsort
    int		compare(const IfShape *shape1, const IfShape *shape2);

    // qsort callback for comparing items
    static int	compareCB(const void *p1, const void *p2)
	{ return curSorter->compare(* (const IfShape **) p1,
				    * (const IfShape **) p2); }
};

#endif /* _IF_SORTER_ */
