/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

//  -*- C++ -*-

/*
 * Copyright (C) 1995-96   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1.1.1 $
 |
 |   Description:
 |  	Import interest list for paste and drop.
 |
 |   Author(s)	: David Mott
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#ifndef  _SO_XT_IMPORT_INTEREST_
#define  _SO_XT_IMPORT_INTEREST_


#include <Inventor/SbPList.h>
#include <Inventor/Xt/SoXtClipboard.h>
#include <X11/Intrinsic.h>

//////////////////////////////////////////////////////////////////////////////
//
//  Class: SoXtImportInterest
//	
//////////////////////////////////////////////////////////////////////////////

class SoXtImportInterest {
  public:
    SoXtImportInterest(Atom _dataType, SoXtClipboardImportCB *f, void *_userData)
	{ dataType = _dataType; cb = f; userData = _userData; }
    ~SoXtImportInterest() {}
    
    Atom		    getDataType() const { return dataType; }
    SoXtClipboardImportCB *  getCallBack() const { return cb; }
    void *		    getUserData() const { return userData; }
  
  private:
    Atom dataType;
    SoXtClipboardImportCB *cb;
    void *userData;
};

//////////////////////////////////////////////////////////////////////////////
//
//  Class: SoXtImportInterestList
//	
//////////////////////////////////////////////////////////////////////////////

class SoXtImportInterestList {
  public: 
    SoXtImportInterestList() {}
    ~SoXtImportInterestList() {}
    
    void			append(Atom dataType,
				    SoXtClipboardImportCB *cb,
				    void *userData);
				    
    int				find(Atom dataType);
    void			remove(int which);
    const SoXtImportInterest *	get(int which);
    Atom			getDataType(int which);
    int				getLength() { return list.getLength(); }
  
  private: 
    SbPList list;
};

#endif /* _SO_XT_IMPORT_INTEREST_ */
