/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

//  -*- C++ -*-

/////////////////////////////////////////////////////////////////////////////
//
// Classes of actions. They are numbered consecutively from 0 and can
// be used as array indices.
//

enum ActionClass {
    GRAPH_CLASS,
    EDIT_CLASS,
    SELECT_CLASS,
    GROUP_CLASS,
    INST_CLASS,
    NODE_CLASS
};

/////////////////////////////////////////////////////////////////////////////
//
// Actions to perform. The XXX_FIRST items are used with the BUT_INDEX
// macro.
//

enum ActionItem {

    GRAPH_FIRST,
    GRAPH_UPDATE,

    EDIT_FIRST,
    EDIT_CUT,
    EDIT_COPY,
    EDIT_PASTE,
    EDIT_PASTE_REF,
    EDIT_DUP,
    EDIT_DUP_REF,
    EDIT_DELETE,

    SELECT_FIRST,
    SELECT_LEFT_SIB,
    SELECT_RIGHT_SIB,
    SELECT_PARENT,
    SELECT_1ST_CHILD,

    GROUP_FIRST,
    GROUP_OPEN,
    GROUP_OPEN_ALL,
    GROUP_CLOSE,
    GROUP_TOGGLE,

    INST_FIRST,
    INST_TOGGLE,
    INST_SWAP,

    NODE_FIRST,
    NODE_CREATE,
    NODE_EDIT,

    NONE 			// For separators in menus, e.g.
};

/////////////////////////////////////////////////////////////////////////////
//
// This macro is used to determine the index of action items within
// their classes. For example, BUT_INDEX(EDIT_DUP, EDIT_FIRST) returns
// the index of the EDIT_DUP action (starting at 0) within the EDIT
// actions.
//

#define ACTION_INDEX(but, first)	((but) - (first) - 1)
