/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

//  -*- C++ -*-

/*
 * Copyright (C) 1990,91   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1.1.1 $
 |
 |   Description:
 |	Internal field converters, created automatically by the
 |      database when hooking up fields of different types.
 |
 |   Author(s)		: Ronen Barzel, Gavin Bell
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#include <Inventor/fields/SoFields.h>

////////////////////////////////////////////////////////////
//
// Converter from various classes to trigger
//
////////////////////////////////////////////////////////////

class SoConvToTrigger : public SoFieldConverter {

    SO_ENGINE_HEADER(SoConvToTrigger);

  public:
    SoField			*input;
    SoEngineOutput		output; /* SoSFTrigger */

  SoINTERNAL public:
    static void			initClass();

    virtual void		evaluate();
    virtual void		inputChanged(SoField *);

  private:
    SoConvToTrigger();
    ~SoConvToTrigger();

    virtual SoField		*getInput(SoType);
    virtual SoEngineOutput	*getOutput(SoType);

    int conversionCase;		// Not used any more

    SoFieldData *myInputData;
};

////////////////////////////////////////////////////////////
//
// Converter for all field conversions except ConvToTrigger.
// It has a big switch inside it with code for each of the
// supported conversions.
//
////////////////////////////////////////////////////////////

class SoBuiltinFieldConverter : public SoFieldConverter {

    SO_ENGINE_HEADER(SoBuiltinFieldConverter);

  public:
    SoField		*input;
    SoEngineOutput	output;

  SoINTERNAL public:
    static void		initClass();
  private:
    SoBuiltinFieldConverter();
    ~SoBuiltinFieldConverter();

    virtual void	evaluate();
    void		doConversion(SoField *outField);

    virtual SoField		*getInput(SoType);
    virtual SoEngineOutput	*getOutput(SoType);

    int inType, outType;

    SoFieldData *myInputData;
    SoEngineOutputData *myOutputData;
};

