/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

/*
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1.1.1 $
 |
 |   Classes	: TsPiece
 |
 |   Author	: Dave Immel
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

//--------------------------- Include ----------------------------------
#include <cstdio>
#include <Inventor/nodes/SoTransform.h>
#include "TsPiece.h"

////////////////////////////////////////////////////////////////////////
//
// Description:
//	Constructor for the TsPiece.
//
// Use: public

TsPiece::TsPiece()

//
////////////////////////////////////////////////////////////////////////
{
    orientation        = 0;
    numBlocks          = 0;
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//	Destructor for TsPiece.
//
// Use: public

TsPiece::~TsPiece()

//
////////////////////////////////////////////////////////////////////////
{
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//	This routine returns the list of positions of the blocks for the
//      piece in its current orientation.  The positions are specified as
//      indices into the standard 27-cube.  This method assumes that space
//      to store the returned indices is passed in.
//
// Use: public

int
TsPiece::getPosition(
        int pos[] )               // (out) the list of positions

// Values Returned:
//      number of blocks in the piece
//
////////////////////////////////////////////////////////////////////////
{
    int i;

    for (i=0; i<numBlocks; i++)
        pos[i] = piecePositionTable[orientation][i];

    return numBlocks;
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//	This routine returns the list of positions of the blocks for the
//      piece in the given orientation.  The positions are specified as
//      indices into the standard 27-cube.  This method assumes that space
//      to store the returned indices is passed in.
//
// Use: public

int
TsPiece::getPosition(
        int newOrientation,
        int pos[])               // (out) the list of positions

// Values Returned:
//      number of blocks in the piece
//
////////////////////////////////////////////////////////////////////////
{
    int i;

    for (i=0; i<numBlocks; i++)
        pos[i] = piecePositionTable[newOrientation][i];

    return numBlocks;
}




