/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

#include "SoMFBitMask.c++"
#include "SoMFBool.c++"
#include "SoMFColor.c++"
#include "SoMFEngine.c++"
#include "SoMFEnum.c++"
#include "SoMFFloat.c++"
#include "SoMFInt32.c++"
#include "SoMFMatrix.c++"
#include "SoMFName.c++"
#include "SoMFNode.c++"
#include "SoMFPath.c++"
#include "SoMFPlane.c++"
#include "SoMFRotation.c++"
#include "SoMFShort.c++"
#include "SoMFString.c++"
#include "SoMFTime.c++"
#include "SoMFUInt32.c++"
#include "SoMFUShort.c++"
#include "SoMFVec2f.c++"
#include "SoMFVec3f.c++"
#include "SoMFVec4f.c++"
#include "SoSFBitMask.c++"
#include "SoSFBool.c++"
#include "SoSFColor.c++"
#include "SoSFEngine.c++"
#include "SoSFEnum.c++"
#include "SoSFFloat.c++"
#include "SoSFImage.c++"
#include "SoSFInt32.c++"
#include "SoSFMatrix.c++"
#include "SoSFName.c++"
#include "SoSFNode.c++"
#include "SoSFPath.c++"
#include "SoSFPlane.c++"
#include "SoSFRotation.c++"
#include "SoSFShort.c++"
#include "SoSFString.c++"
#include "SoSFTime.c++"
#include "SoSFTrigger.c++"
#include "SoSFUInt32.c++"
#include "SoSFUShort.c++"
#include "SoSFVec2f.c++"
#include "SoSFVec3f.c++"
#include "SoSFVec4f.c++"
