/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

#define Arc _lib_nurbs(Arc)
#define ArcSdirSorter _lib_nurbs(ArcSdirSorter)
#define ArcSorter _lib_nurbs(ArcSorter)
#define ArcTdirSorter _lib_nurbs(ArcTdirSorter)
#define ArcTessellator _lib_nurbs(ArcTessellator)
#define Backend _lib_nurbs(Backend)
#define BasicCurveEvaluator _lib_nurbs(BasicCurveEvaluator)
#define BasicSurfaceEvaluator _lib_nurbs(BasicSurfaceEvaluator)
#define BezierArc _lib_nurbs(BezierArc)
#define Bin _lib_nurbs(Bin)
#define Breakpt _lib_nurbs(Breakpt)
#define Buffer _lib_nurbs(Buffer)
#define CachingEvaluator _lib_nurbs(CachingEvaluator)
#define Checker _lib_nurbs(Checker)
#define CoveAndTiler _lib_nurbs(CoveAndTiler)
#define Curve _lib_nurbs(Curve)
#define Curvelist _lib_nurbs(Curvelist)
#define DisplayList _lib_nurbs(DisplayList)
#define Dlnode _lib_nurbs(Dlnode)
#define Edge _lib_nurbs(Edge)
#define Flist _lib_nurbs(Flist)
#define FlistSorter _lib_nurbs(FlistSorter)
#define GridTrimVertex _lib_nurbs(GridTrimVertex)
#define GridVertex _lib_nurbs(GridVertex)
#define Gridline _lib_nurbs(Gridline)
#define Hashtable _lib_nurbs(Hashtable)
#define Hull _lib_nurbs(Hull)
#define Jarcloc _lib_nurbs(Jarcloc)
#define JumpBuffer _lib_nurbs(JumpBuffer)
#define Knotspec _lib_nurbs(Knotspec)
#define Knotvector _lib_nurbs(Knotvector)
#define Mapdesc _lib_nurbs(Mapdesc)
#define Maplist _lib_nurbs(Maplist)
#define Mesher _lib_nurbs(Mesher)
#define Monotonizer _lib_nurbs(Monotonizer)
#define NurbsTessellator _lib_nurbs(NurbsTessellator)
#define O_curve _lib_nurbs(O_curve)
#define O_nurbscurve _lib_nurbs(O_nurbscurve)
#define O_nurbssurface _lib_nurbs(O_nurbssurface)
#define O_pwlcurve _lib_nurbs(O_pwlcurve)
#define O_surface _lib_nurbs(O_surface)
#define O_trim _lib_nurbs(O_trim)
#define Patch _lib_nurbs(Patch)
#define Patchlist _lib_nurbs(Patchlist)
#define Patchspec _lib_nurbs(Patchspec)
#define Pool _lib_nurbs(Pool)
#define PooledObj _lib_nurbs(PooledObj)
#define Property _lib_nurbs(Property)
#define Pspec _lib_nurbs(Pspec)
#define PwlArc _lib_nurbs(PwlArc)
#define PwlArcPool _lib_nurbs(PwlArcPool)
#define Quilt _lib_nurbs(Quilt)
#define Quiltspec _lib_nurbs(Quiltspec)
#define Renderhints _lib_nurbs(Renderhints)
#define Slicer _lib_nurbs(Slicer)
#define Sorter _lib_nurbs(Sorter)
#define Splinespec _lib_nurbs(Splinespec)
#define Subdivider _lib_nurbs(Subdivider)
#define TrimRegion _lib_nurbs(TrimRegion)
#define TrimVertex _lib_nurbs(TrimVertex)
#define TrimVertexPool _lib_nurbs(TrimVertexPool)
#define Trimline _lib_nurbs(Trimline)
#define Uarray _lib_nurbs(Uarray)
#define Varray _lib_nurbs(Varray)
