/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

/*
 * Copyright (C) 1990,91   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1.1.1 $
 |
 |   Classes:
 |      SoV1ShapeKit
 |
 |   Author(s)          : Paul Isaacs
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#include <Inventor/SoDB.h>
#include <Inventor/misc/upgraders/SoV1ShapeKit.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoTextureCoordinateFunction.h>
#include <Inventor/nodes/SoTextureCoordinateBinding.h>
#include <Inventor/nodekits/SoShapeKit.h>

// Define the required type id and name inquiry methods
SO_NODE_SOURCE(SoV1ShapeKit);

// Define the required catalog inquiry methods
SO_V1_SUBKIT_CATALOG_VARS(SoV1ShapeKit);
SO_V1_SUBKIT_CATALOG_METHODS(SoV1ShapeKit);

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Constructor
//
// Use: public

SoV1ShapeKit::SoV1ShapeKit()
//
////////////////////////////////////////////////////////////////////////
{
    SO_NODE_CONSTRUCTOR(SoV1ShapeKit);

    if(SO_NODE_IS_FIRST_INSTANCE() ) {

    // Initialize children catalog and add entries to it
    // These are the macros you use to make a catalog.
    // Use ...INHERIT_CATALOG once, and put it first.
    // Then, use combinations of ...ADD_CATALOG_ENTRY 
    // and ...ADD_CATALOG_LIST_ENTRY.  See SoSubKit.h for more info
    // on syntax of these macros.
    SO_V1_SUBKIT_INHERIT_CATALOG(SoV1ShapeKit, SoV1GroupKit );
    SO_V1_SUBKIT_ADD_CATALOG_ENTRY("materialBinding", SoMaterialBinding,
				    "topSeparator", "childList", TRUE );
    SO_V1_SUBKIT_ADD_CATALOG_ENTRY("normalBinding", SoNormalBinding,
				    "topSeparator", "childList", TRUE );

    SO_V1_SUBKIT_ADD_CATALOG_LIST_ENTRY("textureCoordinateFunctionList", 
  			SoGroup, "topSeparator", "childList", 
                        SoTextureCoordinateFunction, TRUE);

    SO_V1_SUBKIT_ADD_CATALOG_ENTRY("textureCoordinateBinding", 
	    SoTextureCoordinateBinding, "topSeparator", "childList", TRUE );

    SO_V1_SUBKIT_ADD_CATALOG_ABSTRACT_ENTRY("shape", SoShape,
				SoSphere, "topSeparator", "childList", TRUE  );

    }

    
    

    createNodekitPartsList();
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Destructor (necessary since inline destructor is too complex)
//
// Use: public

SoV1ShapeKit::~SoV1ShapeKit()
//
////////////////////////////////////////////////////////////////////////
{
}
SoNode *
SoV1ShapeKit::createNewNode()
{
    return new SoShapeKit;
}


////////////////////////////////////////////////////////////////////////
//
// Description:
//    A virtual function that allows classes to treat parts that fail the
//    tests of 'tryToSetPartInNewNode'
//
// If tryToSetPartInNewNode fails, then this routine is called.
// It will fail for the parts:
// "textureCoordinateFunctionList".
// This part has been changed to "textureCoordinateFunction" which is not
// a list.  This routine will take the last function from the list
// and set it as the part "textureCoordinateFunnction" in the new node.
// It will also print a warning.
//
// Use: public
//
SbBool 
SoV1ShapeKit::dealWithUpgradedPart( SoBaseKit *newNode, SoNode *newPart, 
				    const SbName &newPartName )
//
////////////////////////////////////////////////////////////////////////
{
    // First, try to let base class handle it...
    if ( SoV1BaseKit::dealWithUpgradedPart( newNode, newPart, newPartName ) )
	return TRUE;

    // If the part name is "textureCoordinateFunctionList", get the first 
    // element and set it in the new kit as "textureCoordinateFunction"
    if ( newPartName == "textureCoordinateFunctionList" ) {
	if ( !newPart->isOfType( SoGroup::getClassTypeId() ))
	    return FALSE;

	SoGroup *myGrp = (SoGroup *) newPart;
	if (myGrp->getNumChildren() > 0 ) {
	    SoNode *myNode = myGrp->getChild(myGrp->getNumChildren() - 1);
	    if ( !myNode->isOfType( 
		  SoTextureCoordinateFunction::getClassTypeId() ))
		return FALSE;
	    SoTextureCoordinateFunction *myTexFunct 
		= (SoTextureCoordinateFunction *) myNode;
	    SoDebugError::postWarning("SoV1ShapeKit::dealWithUpgradedPart",
		"the input file contained a textureCoordinateFunctionList part. The new kits have only a textureCoordinateFunction part. Setting textureCoordinateFunction to be the first child in textureCoordinateFunctionList");
	    newNode->setPart("textureCoordinateFunction", myTexFunct );
	}
	return TRUE;
    }

    return FALSE;   // don't know how to do anything yet...
}
