/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

/*
 * Copyright (C) 1990,91   Silicon Graphics, Inc.
 *
 _______________________________________________________________________
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 |
 |   $Revision: 1.1.1.1 $
 |
 |   Classes:
 |	SoViewportRegionElement
 |
 |   Author(s)		: Paul S. Strauss
 |
 ______________  S I L I C O N   G R A P H I C S   I N C .  ____________
 _______________________________________________________________________
 */

#include <Inventor/elements/SoViewportRegionElement.h>

SbViewportRegion *SoViewportRegionElement::emptyViewportRegion;

SO_ELEMENT_SOURCE(SoViewportRegionElement);

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Destructor.
//
// Use: private

SoViewportRegionElement::~SoViewportRegionElement()
//
////////////////////////////////////////////////////////////////////////
{
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Initializes element
//
// Use: public

void
SoViewportRegionElement::init(SoState *)
//
////////////////////////////////////////////////////////////////////////
{
    // Choose a reasonable default window size
    viewportRegion.setWindowSize(100, 100);
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Sets viewport region to given matrix in element accessed from state.
//
// Use: public, static

void
SoViewportRegionElement::set(SoState *state,
			     const SbViewportRegion &vpReg)
//
////////////////////////////////////////////////////////////////////////
{
    SoViewportRegionElement	*elt;

    // Get an instance we can change (pushing if necessary)
    elt = (SoViewportRegionElement *) getElement(state, classStackIndex);

    if (elt != NULL)
	elt->setElt(vpReg);
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Returns viewport region from state.
//
// Use: public, static

const SbViewportRegion &
SoViewportRegionElement::get(SoState *state)
//
////////////////////////////////////////////////////////////////////////
{
    const SoViewportRegionElement *elt;

    elt = (const SoViewportRegionElement *)
	getConstElement(state, classStackIndex);
    if (elt != NULL )
	return elt->viewportRegion;
    else
	return *emptyViewportRegion;
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Returns TRUE if viewport regions are the same -- if they have
//    the same size and origin.
//
// Use: public

SbBool
SoViewportRegionElement::matches(const SoElement *elt) const
//
////////////////////////////////////////////////////////////////////////
{
    return (viewportRegion ==
	    ((const SoViewportRegionElement *) elt)->viewportRegion);
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//     Create a copy of this instance suitable for calling matches()
//     on.
//
// Use: protected

SoElement *
SoViewportRegionElement::copyMatchInfo() const
//
////////////////////////////////////////////////////////////////////////
{
    SoViewportRegionElement *result =
	(SoViewportRegionElement *)getTypeId().createInstance();

    result->viewportRegion = viewportRegion;
    
    return result;
}

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Prints element for debugging.
//
// Use: public
//
////////////////////////////////////////////////////////////////////////

#ifdef DEBUG
void
SoViewportRegionElement::print(FILE *fp) const
{
    SoElement::print(fp);

    fprintf(fp, "\tViewport Region:\n");

    const SbVec2s	&winSize = viewportRegion.getWindowSize();
    const SbVec2f	&vpOrig  = viewportRegion.getViewportOrigin();
    const SbVec2f	&vpSize  = viewportRegion.getViewportSize();

    fprintf(fp, "\t\tWindow   size   = %d, %d\n", winSize[0], winSize[1]);
    fprintf(fp, "\t\tViewport origin = %g, %g\n", vpOrig[0], vpOrig[1]);
    fprintf(fp, "\t\tViewport size   = %g, %g\n", vpSize[0], vpSize[1]);
}
#else  /* DEBUG */
void
SoViewportRegionElement::print(FILE *) const
{
}
#endif /* DEBUG */

////////////////////////////////////////////////////////////////////////
//
// Description:
//    Sets viewport region in element to given viewport region.
//
// Use: protected, virtual

void
SoViewportRegionElement::setElt(const SbViewportRegion &vpReg)
//
////////////////////////////////////////////////////////////////////////
{
    viewportRegion = vpReg;
}
