/*
 *
 *  Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 *  Mountain View, CA  94043, or:
 * 
 *  http://www.sgi.com 
 * 
 *  For further information regarding this notice, see: 
 * 
 *  http://oss.sgi.com/projects/GenInfo/NoticeExplan/
 *
 */

/*
 |   MyGrids.h
 |
 |   Author(s): Paul Isaacs
 |
 */

// This file contains the strings describing 
// the two grids used in noodle
//
// char *myGridBuffer
// char *myGrdiXYBuffer
//

char *myGridBuffer =
{ "#Inventor V2.0 ascii\n\
\n\
Separator {\n\
	PickStyle { style UNPICKABLE }\n\
	LightModel { model BASE_COLOR }\n\
	NormalBinding { value DEFAULT }\n\
# Grid\n\
	BaseColor { rgb 0.2 0.2 0.2 }\n\
	Array {\n\
		numElements1 61\n\
		separation1 .5 0 0\n\
		origin CENTER\n\
\n\
		Coordinate3 { point [ 0 -15 -.00001, 0 15 -.00001 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
	Array {\n\
		numElements1 61\n\
		separation1 0 .5 0\n\
		origin CENTER\n\
\n\
		Coordinate3 { point [ -15 0 -.00001, 15 0 -.00001 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
# Axes\n\
	BaseColor { rgb 0.4 0.0 0.0 }\n\
	    Coordinate3 { point [ -15.0 0.0 0.0, 15.0 0.0 0.0,\n\
			0.0 -15.0 0, 0.0 15.0 0.0 ]\n\
	}\n\
	LineSet { numVertices [ 2, 2 ] }\n\
}"};

char *myProfileGridBuffer =
{ "#Inventor V2.0 ascii\n\
\n\
Separator {\n\
	PickStyle { style UNPICKABLE }\n\
	LightModel { model BASE_COLOR }\n\
	NormalBinding { value DEFAULT }\n\
# Grid\n\
	BaseColor { rgb 0.2 0.2 0.2 }\n\
	Array {\n\
		numElements1 31\n\
		separation1 .5 0 0\n\
		origin FIRST\n\
\n\
		Coordinate3 { point [ 0 -15 -.00001, 0 15 -.00001 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
	Array {\n\
		numElements1 61\n\
		separation1 0 .5 0\n\
		origin CENTER\n\
\n\
		Coordinate3 { point [ 0 0 -.00001, 15 0 -.00001 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
# Axes\n\
	BaseColor { rgb 0.4 0.0 0.0 }\n\
	    Coordinate3 { point [ -15.0 0.0 0.0, 15.0 0.0 0.0,\n\
			0.0 -15.0 0, 0.0 15.0 0.0 ]\n\
	}\n\
	LineSet { numVertices [ 2, 2 ] }\n\
}"};



char *myGridXYBuffer =
{ "#Inventor V2.0 ascii\n\
\n\
Separator {\n\
	PickStyle { style UNPICKABLE }\n\
	LightModel { model BASE_COLOR }\n\
	NormalBinding { value DEFAULT }\n\
# Grid\n\
	BaseColor { rgb 0.2 0.2 0.2 }\n\
	Array {\n\
		numElements1 61\n\
		separation1 .5 0 0\n\
		origin CENTER\n\
\n\
		Coordinate3 { point [ 0 -.00001 -15, 0 -.00001 15 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
	Array {\n\
		numElements1 61\n\
		separation1 0 0 .5\n\
		origin CENTER\n\
\n\
		Coordinate3 { point [ -15 -.00001 0, 15 -.00001 0 ] }\n\
		LineSet { numVertices [ 2 ] }\n\
	}\n\
# Axes\n\
	BaseColor { rgb 0.4 0.0 0.0 }\n\
	DrawStyle { lineWidth 2 }\n\
	Coordinate3 { point [ -15.0 0.0 0.0, 15.0 0.0 0.0,\n\
			0.0 0.0 -15.0, 0.0 0.0 15.0 ]\n\
	}\n\
	LineSet { numVertices [ 2, 2 ] }\n\
}" };
